/*
*  K4Guitune - Simple chromatic tuner for KDE4.
*  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
*
*  This program is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "audiograph.h"

#include "audioprocessor.h"

#include <cmath>

#include <KDE/KLocale>

#include <QPainter>
#include <QResizeEvent>


// ** CUSTOM CONSTANTS ** //
const double    SCOPE_WIDTH     = 0.88;
const double    SCOPE_HEIGHT    = 0.36;


AudioGraph::AudioGraph( QWidget *parent ) : QWidget( parent ),
	m_pixmap( 0 ),
	m_sample( new float[AudioProcessor::PLOT_SIZE] ),
	m_autocorrelation( new float[AudioProcessor::PLOT_SIZE] ),
	m_estimatedFrequency( 0.0 )
{
	memset( m_sample, 0, AudioProcessor::PLOT_SIZE * sizeof(float) );
	memset( m_autocorrelation, 0, AudioProcessor::PLOT_SIZE * sizeof(float) );

	Q_CHECK_PTR( m_sample );
	Q_CHECK_PTR( m_autocorrelation );	
}


AudioGraph::~AudioGraph()
{
	delete[] m_sample;
	delete[] m_autocorrelation;
}


void AudioGraph::setSignalColor( const QColor &signalColor )
{
	m_signalColor = signalColor;
	update( );
}


void AudioGraph::setAutocorrelationColor( const QColor &autocorrelationColor )
{
	m_autocorrelationColor = autocorrelationColor;
	update( );
}


void AudioGraph::setData( const float *sample, const float *autocorrelation, double estimatedFrequency )
{
	Q_CHECK_PTR( sample );
	Q_CHECK_PTR( autocorrelation );

	memcpy( m_sample, sample, AudioProcessor::PLOT_SIZE * sizeof(float) );
	memcpy( m_autocorrelation, autocorrelation, AudioProcessor::PLOT_SIZE * sizeof(float) );
	m_estimatedFrequency = estimatedFrequency;
	update( );
}


void AudioGraph::paintEvent( QPaintEvent *event )
{
	Q_UNUSED( event );

	int width  = (int)(SCOPE_WIDTH * this->width( ));
	int height = (int)(SCOPE_HEIGHT * this->height( )) / 2;

	QPainter painter( this );
	painter.drawPixmap( 0, 0, *m_pixmap );

	painter.translate( (int)(this->width( ) * (1 - SCOPE_WIDTH) / 2) + 1, (int)(this->height( ) * 0.25) );
	drawCurve( painter, m_sample, AudioProcessor::PLOT_SIZE, width - 2, height, m_signalColor );

	painter.translate( 0, (int)(this->height( ) * 0.5) );
	drawCurve( painter, m_autocorrelation, AudioProcessor::PLOT_SIZE, width - 2, height, m_autocorrelationColor );

	if ( (m_estimatedFrequency >= 50.0) && (m_estimatedFrequency <= 2000.0) ) {
		painter.setPen( QPen( Qt::red, 1, Qt::SolidLine ) );
		painter.setRenderHint( QPainter::Antialiasing, true );
		painter.drawLine( QPointF( 50.0 / m_estimatedFrequency * width, -height ),
			QPointF( 50.0 / m_estimatedFrequency * width, height ) );
		painter.setRenderHint( QPainter::Antialiasing, false );
	}
}


void AudioGraph::resizeEvent( QResizeEvent *event )
{
	cachePixmap( event->size( ) );
}


void AudioGraph::drawAxisBox( QPainter &painter, const int width, const int height )
{
	painter.save( );

	painter.setPen( QPen( Qt::white, 1, Qt::DashLine ) );
	painter.drawLine( 0, 0, width - 1, 0 );

	painter.setPen( QPen( Qt::white, 1, Qt::SolidLine ) );
	painter.drawRect( 0, -height - 1, width - 1, 2 * height + 1);

	painter.restore( );
}


void AudioGraph::drawSignalAxis( QPainter &painter, const int width, const int height )
{
	drawAxisBox( painter, width, height );

	painter.save( );

	QFont font = painter.font( );
	font.setPointSize( 7 );
	painter.setFont( font );
	painter.setPen( QPen( Qt::white, 1, Qt::SolidLine ) );

	const int xAxisRange = 20;
	for ( int k = 1 ; k < xAxisRange ; ++k ) {
		int xTick = (int)((k * width) / xAxisRange);
		painter.drawLine( xTick, height, xTick, height - ( (k % 5) == 0 ? 6 : 3 ) );
	}

	QString textLabel;
	for ( int k = 0 ; k <= 6 ; ++k ) {
		textLabel = QString( "%1" ).arg( k / 6.0 * xAxisRange, 0, 'f', 0 );
		painter.drawText( (k * width) / 6 - painter.fontMetrics( ).width( textLabel ) / 2 + 1,
			height + painter.fontMetrics( ).ascent( ) + 6, textLabel );
	}

	painter.restore( );
}


void AudioGraph::drawAutocorrelationAxis( QPainter& painter, const int plotArea_width, const int plotArea_height )
{
	drawAxisBox( painter, plotArea_width, plotArea_height );

	painter.save( );

	QFont font = painter.font( );
	font.setPointSize( 7 );
	painter.setFont( font );
	painter.setPen( QPen( Qt::white, 1, Qt::SolidLine ) );

	int     xTick;
	double  freq = 10.0;

	for ( unsigned int k = 5 ; k <= 20 ; ++k ) {
	if ( (k % 10) == 0 ) {
		freq *= 10;
		xTick = (int)( 50.0 * (double) plotArea_width / freq );
	} else {
		xTick = (int)( 50.0 * (double) plotArea_width / ((k % 10) * freq) );
	}

	if ( freq > 75.0 )
		painter.drawLine( xTick, plotArea_height, xTick, plotArea_height - ( (k % 5) == 0 ? 6 : 3 ) );

	QString textLabel;
		if ( (k % 10) == 0 ) {
			textLabel = QString( "%1" ).arg( (unsigned int) freq );
			painter.drawText( xTick - painter.fontMetrics( ).width( textLabel ) / 2 + 1,
			plotArea_height + painter.fontMetrics( ).ascent( ) + 6, textLabel );
		} else if ( (k % 10) == 5 ) {
			textLabel = QString("%1").arg( (unsigned int)((k % 10) * freq) );
			painter.drawText( xTick - painter.fontMetrics( ).width( QString("%1").arg( (unsigned int)((k % 10) * freq) ) ) / 2 + 1,
			plotArea_height + painter.fontMetrics( ).ascent( ) + 6, textLabel );
		}
	}

	painter.restore( );
}


void AudioGraph::drawCurve( QPainter &painter, const float *data, const unsigned int dataSize,
const int width, const int height, const QColor &color )
{
	Q_CHECK_PTR( data );
	Q_ASSERT( dataSize != 0 );

	painter.setPen( QPen( color, 1, Qt::SolidLine ) );

	float xIntervalStep = (float) width / dataSize;
	float yRange        = height * 0.95;
	for ( unsigned int k = 1 ; k < dataSize ; ++k ) {
		painter.drawLine( QPointF( (k - 1) * xIntervalStep, -data[k - 1] * yRange ),
			QPointF( k * xIntervalStep, -data[k] * yRange ) );
	}
}


void AudioGraph::cachePixmap( const QSize &size )
{
	int width  = (int)(SCOPE_WIDTH * size.width( ));
	int height = (int)(SCOPE_HEIGHT * size.height( )) / 2;

	delete m_pixmap;
	m_pixmap = new QPixmap( size );
	m_pixmap->fill( Qt::black );

	QPainter painter( m_pixmap );
	QFont font = painter.font( );
	font.setPointSize( 8 );
	painter.setFont( font );

	painter.translate( (int)(size.width( ) * (1 - SCOPE_WIDTH) / 2), (int)(size.height( ) * 0.25) );
	drawSignalAxis( painter, width, height );
	painter.setPen( QPen( Qt::white, 1, Qt::SolidLine ) );
	painter.drawText( width / 2 - painter.fontMetrics( ).width( i18n("Audio signal [ms]") ) / 2 + 1,
		- height - painter.fontMetrics( ).descent( ) - 6,
	QString("Audio signal [ms]") );

	painter.translate( 0, (int)(size.height( ) * 0.5) );
	drawAutocorrelationAxis( painter, width, height );
	painter.drawText( width / 2 - painter.fontMetrics( ).width( i18n("Autocorrelation [Hz]") ) / 2 + 1,
		- height - painter.fontMetrics( ).descent( ) - 6,
	QString("Autocorrelation [Hz]") );
}
