/*
*  K4Guitune - Simple chromatic tuner for KDE4.
*  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
*
*  This program is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __AUDIOGRAPH_H_
#define __AUDIOGRAPH_H_

#include <QWidget>


/**
 * This class implements an oscilloscope-like widget.  The signal in the time
 * domain is plot in the top axis; the x-axis has a fixed range of roughly 20
 * milliseconds, while the y-axis is normalized in the range [-1.0, 1.0].  The
 * autocorrelation of the input signal is plotted in the lower axis; the x-axis
 * has a (somehow) logarithmic scale ranging from 50 Hz to 1000 Hz. The peak of
 * the autocorrelation used to detect the frequency of the input signal is
 * indicated by a red line.
 *
 * @short Signal visualization in time domain.
 */

class AudioGraph : public QWidget {
	
	Q_OBJECT


public:
	/*!	@brief Deafult constructor.
	 *	@param[in] parent Handle to the parent widget.
	 */
	AudioGraph( QWidget *parent = 0 );

	/*!	@brief Default destructor.
	 */
	~AudioGraph( );

	/*!	@brief Set the color of the signal graph.
	 *	@param[in] signalColor The color of the signal graph.
	 */
	void setSignalColor( const QColor &signalColor );

	/*!	@brief Set the color of the autocorrelation graph.
	 *	@param[in] autocorrelationColor The color of the autocorrelation graph.
	 */
	void setAutocorrelationColor( const QColor &autocorrelationColor );

	/*!	@brief Set the input and autocorrelation signals displayed in the oscilloscope.
	 *	@param[in] sample The array with the new samples of the input signal to display.
	 *	@param[in] autocorrelation The array with the new samples of the autocorrelation to display.
	 *	@param[in] estimatedFrequency The frequency of the input signal estimated as the first autocorrelation peak.
	*/
	void setData( const float *sample, const float *autocorrelation, double estimatedFrequency );

	
protected:
	/*!	@brief Handle a repaint request.
	 *	@param[in] event The details of the repaint event.
	 */
	virtual void paintEvent( QPaintEvent *event );

	/*!	@brief Handle a resize request.
	 *	@param[in] event The details of the resize event.
	 */
	virtual void resizeEvent( QResizeEvent *event );


private:
	QPixmap					*m_pixmap;							//!< @brief Pixmap used to store the background to reduce the load.
	float					*m_sample;							//!< @brief Buffer used to store time samples for visualization [0, 20] ms.
	float					*m_autocorrelation;					//!< @brief Buffer used to store autocorrelation samples for visualization [50, 1000] Hz.
	QColor					m_signalColor;						//!< @brief The color of the signal graph.
	QColor					m_autocorrelationColor;				//!< @brief The color of the autocorrelation graph.
	double					m_estimatedFrequency;				//!< @brief Value of the estimated frequency.

	/*!	@brief Draws an empty axis box with an x-axis.
	 *	@param[in] painter Reference to the painter object used to draw on screen.
	 *	@param[in] width Width of the plot area.
	 *	@param[in] height Half the height of the plot area (used instead of the height to simplify coding).
	 */
	void drawAxisBox( QPainter &painter, const int width, const int height );

	/*!	@brief Draws an axis box with a linear scale with a fixed range [0, 20] ms.
	 *	@param[in] painter Reference to the painter object used to draw on screen.
	 *	@param[in] width Width of the plot area.
	 *	@param[in] height Half the height of the plot area (used instead of the height to simplify coding).
	 */
	void drawSignalAxis( QPainter &painter, const int width, const int height );

	/*!	@brief Draw an axis box with a reversed logarithmic scale with a fixed range [1000, 50] Hz.
	 *	@param[in] painter Reference to the painter object used to draw on screen
	 *	@param[in] width Width of the plot area
	 *	@param[in] height Half the height of the plot area (used instead of the height to simplify coding)
	 */
	void drawAutocorrelationAxis( QPainter &painter, const int width, const int height );

	/*!	@brief Draws an array of normalized data in the range [-1.0, 1.0].
	 *	@param[in] painter Reference to the painter object used to draw on screen.
	 *	@param[in] data Buffer with the data to draw.
	 *	@param[in] dataSize Size of the data buffer.
	 *	@param[in] width Width of the plot area.
	 *	@param[in] height Half the height of the plot area (used instead of the height to simplify coding).
	 *	@param[in] color The color used to draw the data.
	 */
	void drawCurve( QPainter &painter, const float *data, const unsigned int dataSize,
		const int width, const int height, const QColor &color );

	/*!	@brief Redraw the widget background and cache it to reduce load.
	 *	@param[in] size The size of the widget.
	 */
	void cachePixmap( const QSize &size );
};

#endif // __AUDIOGRAPH_H_
