/*
 *  K4Guitune - Simple chromatic tuner for KDE4.
 *  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "audioprocessor.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <linux/soundcard.h>

#include <QMutex>

#include <KDE/KLocale>

const char		OSS_DEVICE[]			= "/dev/dsp";							// The default audio device for OSS.

const int		MAX_FFT_SIZE			= 2048;									// The size of the read buffer (max FFT window size).
const int		SAMPLES_TO_READ			= 512;									// Number of samples to read each time (min FFT window size).
const int		ZERO_PADDING_FACTOR		= 16;									// Number of times that the FFT is zero-padded to increase frequency resolution.
const int		FFTW_BUFFER_SIZE		= MAX_FFT_SIZE * ZERO_PADDING_FACTOR;	// Size of the internal buffer used to compute FFT.

const double	SIGNAL_THRESHOLD_ON		= 100;									// Value of the threshold above which the processing is activated.
const double	SIGNAL_THRESHOLD_OFF	= 20;									// Value of the threshold below which the input audio signal is deactivated.


AudioProcessor::AudioProcessor( QObject *parent ) : QThread( parent ),
	m_buffer( new short int[MAX_FFT_SIZE] ),
	m_bufferIndex( 0 ),
	m_audioLevel( false ),
	m_plotBuffer( new float[PLOT_SIZE] ),
	m_plotAutocorr( new float[PLOT_SIZE] ),
	m_estimatedFrequency( 0.0 ),
	m_fftwPlanFFT( 0 ),
	m_fftwPlanIFFT( 0 ),
	m_running( false ),
	m_mutex( new QMutex( ) )
{
	m_fftwBufferTime	= (double*) fftw_malloc( FFTW_BUFFER_SIZE * sizeof(double) );
	m_fftwBufferFreq	= (fftw_complex*) fftw_malloc( FFTW_BUFFER_SIZE * sizeof(fftw_complex) );

	memset( m_buffer, 0, MAX_FFT_SIZE * sizeof(short int) );
	memset( m_fftwBufferTime, 0, FFTW_BUFFER_SIZE * sizeof(double) );
	memset( m_fftwBufferFreq, 0, FFTW_BUFFER_SIZE * sizeof(fftw_complex) );
	memset( m_plotBuffer, 0, PLOT_SIZE * sizeof(float) );
	memset( m_plotAutocorr, 0, PLOT_SIZE * sizeof(float) );
	
	Q_CHECK_PTR( m_buffer );
	Q_CHECK_PTR( m_mutex );
	Q_CHECK_PTR( m_fftwBufferTime );
	Q_CHECK_PTR( m_fftwBufferFreq );
	Q_CHECK_PTR( m_plotBuffer );
	Q_CHECK_PTR( m_plotAutocorr );
}


AudioProcessor::~AudioProcessor( )
{
	m_mutex->lock( );
	m_running = false;
	m_mutex->unlock( );

	wait( );
	close( m_ossFd );

	fftw_destroy_plan( m_fftwPlanFFT );
	fftw_destroy_plan( m_fftwPlanIFFT );
	fftw_free( m_fftwBufferTime );
	fftw_free( m_fftwBufferFreq );
	
	delete[] m_buffer;
	delete[] m_plotBuffer;
	delete[] m_plotAutocorr;
}


void AudioProcessor::init( )
{
	m_ossFd = open( OSS_DEVICE, O_RDONLY | O_NONBLOCK );
	if ( m_ossFd == -1 ) {
		emit signalCriticalError( i18n( "Unable to open audio device %1: %2" ).arg( OSS_DEVICE ).arg( strerror( errno ) ) );
	} else {
		
		fcntl( m_ossFd, F_SETFL, O_RDONLY );
		ioctl( m_ossFd, SNDCTL_DSP_SETDUPLEX, 0 );

		int format = AFMT_S16_LE;
		ioctl( m_ossFd, SNDCTL_DSP_SETFMT, &format );

		int stereo = 0;
		ioctl( m_ossFd, SNDCTL_DSP_STEREO, &stereo );
	}
}


void AudioProcessor::setParameters( int sampleRate, int fftSize )
{
	Q_ASSERT( fftSize <= MAX_FFT_SIZE );

	m_mutex->lock( );
	m_running = false;
	m_mutex->unlock( );
	
	wait( );
	
	if ( m_sampleRate != sampleRate ) {
		m_sampleRate = sampleRate;
		ioctl( m_ossFd, SOUND_PCM_WRITE_RATE, &sampleRate );
		ioctl( m_ossFd, SOUND_PCM_READ_RATE, &m_sampleRate );
	}

	if ( m_fftSize != fftSize ) {
		m_fftSize = fftSize;

		m_fftSize = fftSize;
		fftw_destroy_plan( m_fftwPlanFFT );
		fftw_destroy_plan( m_fftwPlanIFFT );

		m_fftwPlanFFT		= fftw_plan_dft_r2c_1d( m_fftSize, m_fftwBufferTime, m_fftwBufferFreq, FFTW_ESTIMATE );							// FFT
		m_fftwPlanIFFT		= fftw_plan_dft_c2r_1d( ZERO_PADDING_FACTOR * m_fftSize, m_fftwBufferFreq, m_fftwBufferTime, FFTW_ESTIMATE );	// IFFT zero-padded
	}

	Q_CHECK_PTR( m_fftwPlanFFT );
	Q_CHECK_PTR( m_fftwPlanIFFT );

	m_running			= true;
	start( );
}


void AudioProcessor::audioData( float *buffer, float *autocorrelation, double &estimatedFrequency ) const
{
	Q_CHECK_PTR( buffer );
	Q_CHECK_PTR( autocorrelation );

	m_mutex->lock( );
	memcpy( buffer, m_plotBuffer, PLOT_SIZE * sizeof(float) );
	memcpy( autocorrelation, m_plotAutocorr, PLOT_SIZE * sizeof(float) );
	estimatedFrequency = m_estimatedFrequency;
	m_mutex->unlock( );
}


void AudioProcessor::run( )
{
	short int buffer[SAMPLES_TO_READ];
	
	forever {
		m_mutex->lock( );
		if ( m_running == false ) {
			m_mutex->unlock( );
			break;
		}
		m_mutex->unlock( );
		
		if ( read( m_ossFd, buffer, SAMPLES_TO_READ * sizeof(short int) ) == (SAMPLES_TO_READ * sizeof(short int)) ) {
			memcpy( m_buffer + m_bufferIndex, buffer, SAMPLES_TO_READ * sizeof(short int) );
			m_bufferIndex += SAMPLES_TO_READ;

			for ( int k = 0 ; k < m_fftSize ; ++k )
				m_fftwBufferTime[k] = (double) m_buffer[(m_bufferIndex - m_fftSize + k) % MAX_FFT_SIZE];
			
			if ( m_bufferIndex >= MAX_FFT_SIZE )
				m_bufferIndex = 0;
		} else {
			emit signalCriticalError( i18n( "Read error on %1: %2" ).arg( OSS_DEVICE ).arg( strerror( errno ) ) );
			return;
		}

		double signalLevel = 0.0;
		if ( m_audioLevel == false ) {
			double sample;
			for ( int k = 0 ; k < m_fftSize ; ++k ) {
				sample = m_fftwBufferTime[k];
				if ( (sample >= SIGNAL_THRESHOLD_ON) || (sample <= -SIGNAL_THRESHOLD_ON) )
					m_audioLevel = true;

				if ( sample > signalLevel )
					signalLevel = sample;
				if ( -sample > signalLevel )
					signalLevel = -sample;
			}
		} else {
			double sample;
			m_audioLevel = false;
			for ( int k = 0 ; k < m_fftSize ; ++k ) {
				sample = m_fftwBufferTime[k];
				if ( (sample >= SIGNAL_THRESHOLD_OFF) || (sample <= -SIGNAL_THRESHOLD_OFF) )
					m_audioLevel = true;
				
				if ( sample > signalLevel )
					signalLevel = sample;
				if ( -sample > signalLevel )
					signalLevel = -sample;
			}
		}
	
		if ( m_audioLevel == false ) {
			memset( m_plotBuffer, 0, PLOT_SIZE * sizeof(float) );
			memset( m_plotAutocorr, 0, PLOT_SIZE * sizeof(float) );
			m_estimatedFrequency = 0.0;
		} else {
			m_mutex->lock( );
 			unsigned int 	fftwTimeDownsample 	= (m_sampleRate == 22050) ? 2 : 1;
			double 			scaleFactor			= (signalLevel <= 4096.0) ? 4096.0 : signalLevel;

			int triggerIndex = m_fftSize - fftwTimeDownsample * PLOT_SIZE - 1;
			while ( triggerIndex > 0 ) {
				if ( (m_fftwBufferTime[triggerIndex-1] < 0.0) && (m_fftwBufferTime[triggerIndex] >= 0.0) )
					break;
				triggerIndex--;
			}
			Q_ASSERT( (int)(PLOT_SIZE * fftwTimeDownsample + triggerIndex) < m_fftSize );
			
			for ( int k = 0 ; k < PLOT_SIZE ; ++k ) {
				m_plotBuffer[k] = (float)( m_fftwBufferTime[k * fftwTimeDownsample + triggerIndex] / scaleFactor );
 			}
			m_mutex->unlock( );
			
			// ** COMPUTE THE AUTOCORRELATION ** //
			// compute the FFT of the input signal
			fftw_execute( m_fftwPlanFFT );

			/* Compute the transform of the autocorrelation given in time domain by
			*           k=-N
			*    r[t] = sum( x[k] * x[t-k] )
			*            N
			* or in the frequency domain (for a real signal) by
			*    R[f] = X[f] * X[f]' = |X[f]|^2 = Re(X[f])^2 + Im(X[f])^2
			* When computing the FFT with fftw_plan_dft_r2c_1d there are only N/2
			* significant samples, so |.|^2 is computed for m_fftSize/2 samples only
			*/

			// compute |.|^2 of the signal
			for( int k = 0 ; k < (m_fftSize / 2 + 1) ; ++k ) {
				m_fftwBufferFreq[k][0] = (m_fftwBufferFreq[k][0] * m_fftwBufferFreq[k][0]) + (m_fftwBufferFreq[k][1] * m_fftwBufferFreq[k][1]);
				m_fftwBufferFreq[k][1] = 0.0;
			}

			// pad the FFT with zeros to increase resolution
			memset( &(m_fftwBufferFreq[m_fftSize/ 2 + 1][0]), 0, ( (ZERO_PADDING_FACTOR - 1) * m_fftSize + m_fftSize/ 2 - 1) * sizeof(fftw_complex) );

			// compute the IFFT to obtain the autocorrelation in time domain
			fftw_execute( m_fftwPlanIFFT );

			// ** FIND THE MAXIMUM OF THE AUTOCORRELATION (REJECTING THE FIRST PEAK) ** //
			// search for a minimum in the autocorrelation to reject the peak centered around 0
			int l = 0;
			while ( (l < (ZERO_PADDING_FACTOR / 2 * m_fftSize + 1)) && ( (m_fftwBufferTime[l+1] < m_fftwBufferTime[l]) || (m_fftwBufferTime[l+1] > 0.0) ) )
				++l;

			// search for the maximum
			double	maxAutocorr			= 0.0;
			int		maxAutocorrIndex	= 0;
			while ( l < ( ZERO_PADDING_FACTOR / 2 * m_fftSize + 1) ) {
				if ( m_fftwBufferTime[l] > maxAutocorr ) {
					maxAutocorr			= m_fftwBufferTime[l];
					maxAutocorrIndex	= l;
				}
				++l;
			}

			// compute the frequency of the maximum considering the padding factor
			m_mutex->lock( );
			if ( maxAutocorr != 0.0 )
				m_estimatedFrequency = ( ZERO_PADDING_FACTOR * m_sampleRate / (double) maxAutocorrIndex );
			else
				m_estimatedFrequency = 0.0;
			
			int fftwFreqDownsample = ((m_sampleRate == 22050) ? 2 : 1) * ZERO_PADDING_FACTOR;
			Q_ASSERT( (PLOT_SIZE * fftwFreqDownsample) < FFTW_BUFFER_SIZE );
			for ( int k = 0 ; k < PLOT_SIZE ; ++k ) {
				m_plotAutocorr[k] = (float)( m_fftwBufferTime[k * fftwFreqDownsample] / m_fftwBufferTime[0] );
			}
			m_mutex->unlock( );
		}
	}
}
