/*
 *  K4Guitune - Simple chromatic tuner for KDE4.
 *  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __AUDIOPROCESSOR_H_
#define __AUDIOPROCESSOR_H_

#include <QThread>

#include <fftw3.h>

class QMutex;


/**
 * This class processes the audio stream to find its frequency.  The audio
 * stream is acquired through the OSS API using which simply.  The frequency
 * detection algorithm is based on the identification of the first peak in the
 * autocorrelation of the signal, which is computed as the inverse FFT of the
 * power spectral density of the signal (the squared module of the signal FFT).
 * The FFT is computed using the FFTW3 library and prior to the inverse
 * transform the signal is zero-padded to increase the resolution of the
 * autocorrelation in order to have a better frequency identification.
 *
 * @short Frequency detection class.
 */

class AudioProcessor : public QThread
{
	Q_OBJECT

public:
	static const int		PLOT_SIZE = 220;				//!< @brief The size of the plot buffers.
	
	/*!	@brief Default constructor.
	 *	@param[in] parent Handle to the parent widget.
	 */
	AudioProcessor( QObject *parent = 0 );

	/*!	@brief Default destructor.
	 */
	~AudioProcessor( );

	/*!	@brief Initialize audio input.
	 */
	void init( );

	/*!	@brief Set the paremeter of the frequency detection algorithm.
	 *	@param[in] sampleRate Sample rate of the input stream.
	 *	@param[in] fftSize The size of audio buffer on which to compute the FFT.
	 */
	void setParameters( int sampleRate, int fftSize );
	
	/*!	@brief Get the audio samples, the computed autocorrelation and the estimated frequency.
	 *	@param[out] buffer The buffer filled with acquired samples.
	 *	@param[out] autocorrelation The buffer filled with computed autocorrelation.
	 *	@param[out] estimatedFrequency The estimated frequency for the buffers.
	 */
	void audioData( float *buffer, float *autocorrelation, double &estimatedFrequency ) const;


signals:
	/*!	@brief Signal a critical error in the application.
	 *	@param[in] error A text describing the error.
	 */
	void signalCriticalError( const QString &error );


protected:
	/*!	@brief The working function of the thread.
	 */
	void run( );


private:
	int						m_ossFd;      					//!< @brief Handle of the /dev/dsp file.
	int						m_sampleRate;					//!< @brief The actual sample rate of the device.
	int						m_fftSize;						//!< @brief Size of the FFT window.
	
	short int				*m_buffer;						//!< @brief The audio buffer that stores the input signal.
	int						m_bufferIndex;					//!< @brief Index of the first empty position in the buffer.
	bool					m_audioLevel;					//!< @brief Whether or not the input level is high enough.
	
	float					*m_plotBuffer;					//!< @brief The audio signal to plot.
	float					*m_plotAutocorr;				//!< @brief The autocorrelation to plot.
	double					m_estimatedFrequency;			//!< @brief The estimated frequency of the signal.
	
	double					*m_fftwBufferTime;				//!< @brief Support buffer used to store signals in the time domain.
	fftw_complex			*m_fftwBufferFreq;				//!< @brief Support buffer used to store signals in the frequency domain.
	fftw_plan				m_fftwPlanFFT;					//!< @brief Plan to compute the FFT of a given signal.
	fftw_plan				m_fftwPlanIFFT;					//!< @brief Plan to compute the IFFT of a given signal (with additional zero-padding).
	
	bool					m_running;						//!< @brief Whether the thread is acquiring samples.
	QMutex					*m_mutex;						//!< @brief Mutex to grant shared access to data.
};

#endif // __AUDIOPROCESSOR_H_
