/*
 *  K4Guitune - Simple chromatic tuner for KDE4.
 *  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "k4guituneview.h"

#include "audioprocessor.h"
#include "settings.h"

#include <KDE/KLocale>
#include <KMessageBox>

#include <QTimer>


K4GuituneView::K4GuituneView( ) :
	m_audioProcessor( new AudioProcessor( ) ),
	m_refreshTimer( new QTimer( ) )
{
	ui_k4guituneview_base.setupUi( this );
	settingsChanged( );

	setAutoFillBackground( true );
	
	connect( m_audioProcessor, SIGNAL(signalCriticalError(const QString&)),
		this, SLOT(showErrorMessage( const QString&)) );
	m_audioProcessor->init( );
	
	connect( m_refreshTimer, SIGNAL(timeout( )),
		this, SLOT(refreshGui( )) );
	m_refreshTimer->start( 50 );
}

K4GuituneView::~K4GuituneView()
{
	delete m_audioProcessor;
}


void K4GuituneView::refreshGui( )
{
	float audioBuffer[AudioProcessor::PLOT_SIZE];
	float autocorrBuffer[AudioProcessor::PLOT_SIZE];
	
	double estimatedFrequency;
	double estimatedPitch;
	m_audioProcessor->audioData( audioBuffer, autocorrBuffer, estimatedFrequency );
	ui_k4guituneview_base.audioGraph->setData( audioBuffer, autocorrBuffer, estimatedFrequency );
	estimatedPitch = ui_k4guituneview_base.pitchDetector->detectPitch( estimatedFrequency );

	if ( estimatedFrequency < 1e-5 ) {
		ui_k4guituneview_base.lcdNumber_frequency->display( "-.--" );
		ui_k4guituneview_base.lcdNumber_pitch->display( "-.--" );
	} else {
 		ui_k4guituneview_base.lcdNumber_frequency->display( QString::number( estimatedFrequency, 'f', 2 ) );
		ui_k4guituneview_base.lcdNumber_pitch->display( QString::number( estimatedPitch, 'f', 2 ) );
	}
}


void K4GuituneView::settingsChanged( )
{
	int sampleRate;
	switch ( Settings::sampleFrequency( ) ) {
		default:
		case Settings::SAMPLE_FREQ_11025_HZ:	sampleRate = 11025;	break;
		case Settings::SAMPLE_FREQ_22050_HZ:	sampleRate = 22050;	break;
	}
	
	int fftSize;
	switch ( Settings::fftSize( ) ) {
		default:
		case Settings::FFT_SIZE_512:	fftSize = 512;	break;
		case Settings::FFT_SIZE_1024:	fftSize = 1024;	break;
		case Settings::FFT_SIZE_2048:	fftSize = 2048;	break;
	}
	m_audioProcessor->setParameters( sampleRate, fftSize );
	
	double rootFrequency;
	switch ( Settings::rootFrequency( ) ) {
		default:
		case Settings::CONCERT_440:		rootFrequency = 440.00;	break;
		case Settings::FRENCH_435:		rootFrequency = 435.00;	break;
		case Settings::SCIENTIFIC_430:	rootFrequency = 430.54;	break;
	}
	ui_k4guituneview_base.pitchDetector->setParameters( Settings::scaleType( ), Settings::tuningMode( ), rootFrequency );
	
	ui_k4guituneview_base.audioGraph->setSignalColor( Settings::signalColor( ) );
	ui_k4guituneview_base.audioGraph->setAutocorrelationColor( Settings::autoCorrelationColor( ) );
	
	emit signalChangeStatusbar( i18n("Settings changed") );
}


void K4GuituneView::showErrorMessage( const QString &error )
{
	KMessageBox::error( this, error, i18n("K4Guitune") );
	QTimer::singleShot( 0, qApp, SLOT(quit( )) );
}

#include "k4guituneview.moc"
