/*
*  K4Guitune - Simple chromatic tuner for KDE4.
*  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
*
*  This program is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pitchdetector.h"

#include "settings.h"

#include <cmath>

#include <KDE/KLocale>

#include <QMouseEvent>
#include <QPainter>
#include <QResizeEvent>


const QString FLAT  = QChar(0x266D);
const QString SHARP = QChar(0x266F);

const QString NoteLabel[6][12] = {
	{  "A", "B" + FLAT  , "B" , "C" , "D" + FLAT  , "D" , "E" + FLAT  , "E",  "F" , "G" + FLAT  , "G"  , "A" + FLAT    },   /* English  */
	{  "A", "A" + SHARP , "B" , "C" , "C" + SHARP , "D" , "D" + SHARP , "E",  "F" , "F" + SHARP , "G"  , "G" + SHARP   },   /* English alternate */
	{  "A", "B"         , "H" , "C" , "D" + FLAT  , "D" , "E" + FLAT  , "E",  "F" , "G" + FLAT  , "G"  , "A" + FLAT    },   /* German */
	{  "A", "B"         , "H" , "C" , "C" + SHARP , "D" , "D" + SHARP , "E",  "F" , "F" + SHARP , "G"  , "G" + SHARP   },   /* German alternate */
	{ "La", "Si" + FLAT , "Si", "Do", "Re" + FLAT , "Re", "Mi" + FLAT , "Mi", "Fa", "Sol" + FLAT, "Sol", "La" + FLAT   },   /* Italian */
	{ "La", "La" + SHARP, "Si", "Do", "Do" + SHARP, "Re", "Re" + SHARP, "Mi", "Fa", "Fa" + SHARP, "Sol", "Sol" + SHARP }    /* Italian alternate */
};

const QString JustIntonationIntervals[24] = {
	"1/1"	/* Same note */,					"25/24"	/* Minor semitone */,				"16/15"	/* Major semitone */,
	"9/8"	/* Major second */,					"7/6"	/* Just Intonation (10th) */,		"75/64"	/* Augmented second */,
	"6/5"	/* Minor third */,					"5/4"	/* Major third */,					"32/25"	/* Diminished fourth */,
	"125/96"/* Augmented third */,				"4/3"	/* Perfect forth */,				"25/18"	/* Augmented forth */,
	"45/32"	/* Tritone */,						"36/25"	/* Diminished fifth */,				"3/2"	/* Perfect fifth */,
	"25/16"	/* Augmented fifth */,				"8/5"	/* Minor sixth */,					"5/3"	/* Major sixth */,
	"125/72"/* Augmented sixth */,				"16/9"	/* Pythagorean minor seventh */,    "9/5"	/* Minor seventh */,
	"15/8"  /* Major seventh */,				"48/25"	/* Diminished octave */,			"125/64"/* Augmented seventh */
};

const double	M_LOG10_2			= std::log10( 2.0 );

const int		HORIZONTAL_PADDING	= 15;
const int		BAR_HEIGHT			= 8;
const int		CURSOR_WIDTH		= 8;
const double	ACCEPTED_DEVIATION	= 0.025;



PitchDetector::PitchDetector( QWidget *parent ) : QWidget( parent ),
	m_rootFrequency( 440.0 ),
	m_scaleType( 0 ),
	m_tuningMode( Settings::EQUAL_TEMPERAMENT ),
	m_pitchEstimate( 0.0 ),
	m_pitchIndex( 0 ),
	m_pitchOctave( 0 ),
	m_pitchDeviation( 0.0 ),
	m_justIntonationKeyTone( 0 ),
	m_pixmap( 0 ),
	m_showCursor( false )
{
	for ( int k = 0 ; k < 24 ; ++k ) {
		QStringList fraction = JustIntonationIntervals[k].split( '/' );
		m_justIntonationRatios[k] = std::log10( fraction.at( 0 ).toDouble( ) / fraction.at( 1 ).toDouble( ) ) / M_LOG10_2;
	}
}


void PitchDetector::setParameters( int scaleType, int tuningMode, double rootFrequency )
{
	Q_ASSERT( scaleType < 6 );
	Q_ASSERT( tuningMode < 2 );

	m_scaleType = scaleType;
	m_tuningMode = tuningMode;
	m_rootFrequency = rootFrequency;
	
	cachePixmap( size( ) );
	update( );
}


double PitchDetector::detectPitch( double estimatedFrequency )
{
	if ( estimatedFrequency > 1e-5 ) {
		switch ( m_tuningMode ) {
			case Settings::EQUAL_TEMPERAMENT:
				{
					double	semitoneDistance	= std::log10( estimatedFrequency / m_rootFrequency ) * 12.0 / M_LOG10_2;
					int		semitoneRounded		= (int) ( semitoneDistance + ((semitoneDistance > 0.5) ? 0.5 : -0.5) );

					m_pitchOctave				= (int) std::floor( semitoneRounded / 12.0 );
					m_pitchIndex				= (semitoneRounded - 12 * m_pitchOctave + 120) % 12;
					m_pitchDeviation			= (semitoneRounded - semitoneDistance);
					m_pitchEstimate				= m_rootFrequency * std::pow( 2.0, semitoneRounded / 12.0 );
					m_showCursor				= true;
				}
				break;
				
			case Settings::JUST_INTONATION:
				{
					double	keyFreq				= m_rootFrequency * std::pow( 2.0, m_justIntonationKeyTone / 12.0 );
					double	keyDistance			= std::log10( estimatedFrequency / keyFreq ) / M_LOG10_2;
					double	baseDistance		= std::fmod( keyDistance + 10.0, 1.0 );
					
					m_pitchOctave				= (int) std::floor( keyDistance );
					
					m_pitchDeviation			= 1.0;
					for ( int k = 0 ; k < 24 ; ++k ) {
						double deviation = std::fabs( m_justIntonationRatios[k] - baseDistance );
						if ( deviation < m_pitchDeviation ) {
							m_pitchDeviation	= deviation;
							m_pitchIndex		= k;
						}
					}

					if ( std::fabs(baseDistance - 1.0) <= m_pitchDeviation ) {
						m_pitchDeviation	= std::fabs(baseDistance - 1.0);
						m_pitchIndex		= 0;
						m_pitchOctave++;
					}
					m_pitchEstimate				= keyFreq * std::pow( 2.0, m_pitchOctave + m_justIntonationRatios[m_pitchIndex] );
				}
				break;
		}
		
		m_showCursor = true;
	} else {
		m_pitchEstimate				= 0.0;
		m_showCursor				= false;
	}

	update( );
	return m_pitchEstimate;
}


void PitchDetector::paintEvent( QPaintEvent *event )
{
	Q_UNUSED( event );
	Q_CHECK_PTR( m_pixmap );

	QPainter	painter( this );
	int			width	= m_pixmap->width( ) - 2 * HORIZONTAL_PADDING;
	int			height	= m_pixmap->height( );
	QFont		font	= painter.font( );

	painter.drawPixmap( 0, 0, *m_pixmap );

	if ( m_showCursor == true ) {
		font.setPointSize( 14 );
		font.setBold( true );
		painter.setFont( font );
		int fontDescent = painter.fontMetrics( ).descent( );
		
		painter.translate( HORIZONTAL_PADDING, BAR_HEIGHT + font.pointSize( ) + fontDescent + 5 );
		painter.setPen( QPen( QColor( "#00AA00" ), 1, Qt::SolidLine ) );
		
		switch ( m_tuningMode ) {
			default:
			case Settings::EQUAL_TEMPERAMENT:
				{
					Q_ASSERT( m_pitchIndex >= 0 );
					Q_ASSERT( m_pitchIndex < 12 );

					int xTick = (int)((2 * m_pitchIndex + 1) * width / 24.0);
					painter.drawText( xTick - (painter.fontMetrics( ).width( NoteLabel[m_scaleType][m_pitchIndex] ) / 2 + 1), - BAR_HEIGHT - 5 - fontDescent, NoteLabel[m_scaleType][m_pitchIndex] );
					
					font.setPointSize( 7 );
					font.setBold( false );
					painter.setFont( font );
					int fontAscent = painter.fontMetrics( ).ascent( ) + 3;
					int fontHeight = painter.fontMetrics( ).height( );

					if ( (m_pitchOctave >= -3 ) && (m_pitchOctave <= 0 ) ) {
						QString	frequencyLabel	= QString( "%1" ).arg( m_pitchEstimate, 5, 'f', 2 );
						painter.drawText( xTick - (painter.fontMetrics( ).width( frequencyLabel ) / 2 + 1), (3 + m_pitchOctave) * fontHeight + fontAscent + BAR_HEIGHT + 5,
							frequencyLabel );
					}
				}
				break;

			case Settings::JUST_INTONATION:
				{
					Q_ASSERT( m_pitchIndex >= 0 );
					Q_ASSERT( m_pitchIndex < 24 );

					font.setPointSize( 7 );
					font.setBold( false );
					painter.setFont( font );
					int fontAscent = painter.fontMetrics( ).ascent( ) + 3;
					int fontHeight = painter.fontMetrics( ).height( );
					int fontOffset = 2 * BAR_HEIGHT + font.pointSize( ) + fontDescent + 8 + fontHeight;

					int xTick;
					if ( m_pitchIndex == 0 ) {
						xTick = (int)((2 * m_justIntonationKeyTone + 1) * width / 24.0);
						painter.drawLine( xTick, BAR_HEIGHT + 1, xTick, height - fontOffset - 2 );
						painter.drawText( xTick - painter.fontMetrics( ).width( i18n("Key") )/2 + 1,
							height - fontOffset + fontAscent, i18n("Key") );
					} else {
						xTick = (int)( width * (m_justIntonationRatios[m_pitchIndex] + (2 * m_justIntonationKeyTone + 1) / 24.0) ) % width;
						painter.drawLine( xTick, BAR_HEIGHT, xTick, height - fontOffset - (fontHeight + 3) * (2 - (m_pitchIndex % 3)) - 2 );
						painter.drawText( xTick - painter.fontMetrics( ).width( JustIntonationIntervals[m_pitchIndex] )/2 + 1,
							height - fontOffset - (fontHeight + 3) * (2 - (m_pitchIndex % 3)) + fontAscent,
							JustIntonationIntervals[m_pitchIndex] );
					}
				}
				break;
		}

		painter.setPen( QPen( palette( ).dark( ), 1, Qt::SolidLine, Qt::SquareCap, Qt::MiterJoin ) );
		
		int xCursor;
		switch ( m_tuningMode ) {
			default:
			case Settings::EQUAL_TEMPERAMENT:
				xCursor = (int)( width / 12.0 * (m_pitchIndex + m_pitchDeviation + 0.5) );
				break;
			
			case Settings::JUST_INTONATION:
				xCursor = 1 + (int)( width * (m_justIntonationRatios[m_pitchIndex] + (m_justIntonationKeyTone + 0.5) / 12.0) ) % width;
				break;
		}

		if ( m_pitchDeviation < ACCEPTED_DEVIATION ) {
			painter.setRenderHint( QPainter::Antialiasing, true );

			QPolygon rightArrow;
			rightArrow << QPoint( xCursor - CURSOR_WIDTH, -BAR_HEIGHT + 1 )
				<< QPoint( xCursor - CURSOR_WIDTH, BAR_HEIGHT )
				<< QPoint( xCursor, 0 )
				<< QPoint( xCursor - CURSOR_WIDTH, -BAR_HEIGHT + 1 );

			QPainterPath path;
			path.addPolygon( rightArrow );

			painter.fillPath( path, QColor( "#00AA00" ) );
			painter.drawPath( path );

			painter.setRenderHint( QPainter::Antialiasing, false );
		}
		
		if ( m_pitchDeviation > -ACCEPTED_DEVIATION ) {
			painter.setRenderHint( QPainter::Antialiasing, true );

			QPolygon leftArrow;
			leftArrow << QPoint( xCursor + CURSOR_WIDTH - 1, -BAR_HEIGHT + 1 )
			<< QPoint( xCursor + CURSOR_WIDTH - 1, BAR_HEIGHT )
			<< QPoint( xCursor - 1, 0 )
			<< QPoint( xCursor + CURSOR_WIDTH - 1, -BAR_HEIGHT + 1 );

			QPainterPath path;
			path.addPolygon( leftArrow );

			painter.fillPath( path, QColor( "#00AA00" ) );
			painter.drawPath( path );

			painter.setRenderHint( QPainter::Antialiasing, false );
		}
	}
}


void PitchDetector::resizeEvent( QResizeEvent *event )
{
	cachePixmap( event->size( ) );
}


void PitchDetector::cachePixmap( const QSize &size )
{
	delete m_pixmap;
	m_pixmap = new QPixmap( size );
	m_pixmap->fill( Qt::transparent );

	int width	= size.width( ) - 2 * HORIZONTAL_PADDING;
	int height	= size.height( );

	QPainter painter( m_pixmap );

	QFont font = painter.font( );
	font.setPointSize( 14 );
	font.setBold( true );
	painter.setFont( font );
	int fontDescent = painter.fontMetrics( ).descent( );

	painter.translate( HORIZONTAL_PADDING, BAR_HEIGHT + font.pointSize( ) + fontDescent + 5 );
	painter.setPen( QPen( palette( ).dark( ), 1, Qt::SolidLine ) );
	painter.drawRect( 0, -BAR_HEIGHT, width - 1, 2 * BAR_HEIGHT );
	painter.fillRect( 1, -BAR_HEIGHT + 1, width - 2, 2 * BAR_HEIGHT - 1, palette( ).light( ) );

	int xTick;
	int xTickHeight;
	for ( unsigned int k = 0 ; k <= 120 ; ++k ) {
		xTick = (int)((width - 1) * k / 120.0);

		if ( ((k+5) % 10) == 0 )
			xTickHeight = 6;
		else if ( ((k+5) % 5) == 0 )
			xTickHeight = 4;
		else
			xTickHeight = 2;

		painter.drawLine( xTick, -BAR_HEIGHT, xTick, -BAR_HEIGHT - xTickHeight );
		painter.drawLine( xTick,  BAR_HEIGHT, xTick,  BAR_HEIGHT + xTickHeight );
	}

	painter.setPen( QPen( palette( ).text( ), 1, Qt::SolidLine ) );
		for ( unsigned int k = 0 ; k < 12 ; ++k ) {
			xTick = (int)((2*k + 1) * width / 24.0);
			painter.drawText( xTick - (painter.fontMetrics( ).width( NoteLabel[m_scaleType][k] ) / 2 + 1), - BAR_HEIGHT - 5 - fontDescent, NoteLabel[m_scaleType][k] );
	}

	font.setPointSize( 7 );
	font.setBold( false );
	painter.setFont( font );
	int fontAscent = painter.fontMetrics( ).ascent( ) + 3;
	int fontHeight = painter.fontMetrics( ).height( );
	int fontOffset = 2 * BAR_HEIGHT + font.pointSize( ) + fontDescent + 8 + fontHeight;
	
	switch ( m_tuningMode ) {
		case Settings::EQUAL_TEMPERAMENT:
			for ( unsigned int k = 0 ; k < 12 ; ++k ) {
				xTick = (int)((2*k + 1) * width / 24.0);
				QString frequencyLabel;
				double  frequencyValue = m_rootFrequency * std::pow( 2, k / 12.0 );
				for ( int r = 3 ; r >= 0 ; --r ) {
					frequencyLabel = QString( "%1" ).arg( frequencyValue, 5, 'f', 2 );
					painter.drawText( xTick - (painter.fontMetrics( ).width( frequencyLabel ) / 2 + 1), r * fontHeight + fontAscent + BAR_HEIGHT + 5,
						frequencyLabel );
					frequencyValue /= 2;
				}
			}
			break;

		case Settings::JUST_INTONATION:
			xTick = (int)((2 * m_justIntonationKeyTone + 1) * width / 24.0);
			painter.drawLine( xTick, BAR_HEIGHT + 1, xTick, height - fontOffset - 2 );
			painter.drawText( xTick - painter.fontMetrics( ).width( i18n("Key") )/2 + 1,
				height - fontOffset + fontAscent, i18n("Key") );
			
			for ( int k = 1 ; k < 24 ; ++k ) {
				xTick = (int)( width * (m_justIntonationRatios[k] + (2 * m_justIntonationKeyTone + 1) / 24.0) ) % width;
	
				painter.drawLine( xTick, BAR_HEIGHT, xTick, height - fontOffset - (fontHeight + 3) * (2 - (k % 3)) - 2 );
				painter.setPen( QPen( palette( ).text( ), 1 ) );
				painter.drawText( xTick - painter.fontMetrics( ).width( JustIntonationIntervals[k] )/2 + 1,
					height - fontOffset - (fontHeight + 3) * (2 - (k % 3)) + fontAscent,
					JustIntonationIntervals[k] );
			}
			break;
	}
}


void PitchDetector::mousePressEvent( QMouseEvent *event )
{
	m_justIntonationKeyTone = (int)( event->x( ) * 12.0 / width( ) );
	cachePixmap( size( ) );
}
