/*
*  K4Guitune - Simple chromatic tuner for KDE4.
*  Copyright (C) 2009-2010  William Spinelli <wylliam@tiscali.it>
*
*  This program is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//! Pitch detection algorithm and note scale visualization.
/*!
* This class implements a simple pitch detection algorithm to
* identify the note corresponding to the frequency estimated as
* the first peak of the autocorrelation function.
* A linear note scale is displayed using the chosen musical
* notation. A moving cursor gives a rough indication of the
* detected note. The identified note is highlighted and
* when the pitch deviation is smaller than 2.5% a red square
* is drawn around the note label.
*/

#ifndef __PITCHDETECTOR_H_
#define __PITCHDETECTOR_H_

#include <QWidget>


/**
 * This class implements a simple pitch detection algorithm to identify the
 * note corresponding to the frequency estimated.  A linear note scale is
 * displayed using the chosen musical notation.  A moving cursor gives a rough
 * indication of the detected note.
 *
 * @short Pitch detector bar.
 */

class PitchDetector : public QWidget
{
	Q_OBJECT


public:
	/*!	@brief Default constructor.
	 *	@param[in] parent Handle to the parent widget.
	 */
	PitchDetector( QWidget *parent = 0 );

	/*!	@brief Set the tuning notation used to display note names.
	 *	@param[in] scaleType The index of the note names to display.
	 *	@param[in] tuningMode The index tuning mode used to identify frequences.
	 *	@param[in] rootFrequency The root frequency of A4 used for pitch detection.
	 */
	void setParameters( int scaleType, int tuningMode, double rootFrequency );

	/*!	@brief Estimate the pitch associated to the estimated frequency.
	 *	@param[in] estimatedFrequency The estimated frequency used to detect the current pitch.
	 *	@return The detected pitch
	 */
	double detectPitch( double estimatedFrequency );


protected:
	/*!	@brief Handle a repaint request.
	 *	@param[in] event The details of the repaint event.
	 */
	virtual void paintEvent( QPaintEvent *event );

	/*!	@brief Handle a resize request.
	 *	@param[in] event The details of the resize event.
	 */
	virtual void resizeEvent( QResizeEvent *event );

	/*!	@brief Handle a mouse press event.
	 *	@param[in] event The details of the mouse event.
	 */
	virtual void mousePressEvent( QMouseEvent *event );


private:
	double					m_rootFrequency;				//!< @brief Fundamental frequency used as a reference for A4.
	int						m_scaleType;					//!< @brief Musical notation used for the note names.
	int						m_tuningMode;					//!< @brief Tuning mode used to identify frequences.
	
	double					m_pitchEstimate;				//!< @brief The estimated pitch.
	int						m_pitchIndex;					//!< @brief Estimate pitch index with respect to A4.
	int						m_pitchOctave;					//!< @brief Estimated pitch octave with respect to A4.
	double					m_pitchDeviation;				//!< @brief Pitch deviation from the nearest note in the range [-0.5, +0.5].

	double					m_justIntonationRatios[24];		//!< @brief The numerical value of just intonation ratios.
	int						m_justIntonationKeyTone;		//!< @brief The index of the key not in just intonation.
	
	QPixmap					*m_pixmap;						//!< @brief Pixmap used to store the background to reduce the computational load.
	bool					m_showCursor;					//!< @brief Control whether to show the cursor or not.

	/*! @brief Redraw the widget background and cache it to reduce load.
	*  @param[in] size The size of the widget.
	*/
	void cachePixmap( const QSize &size );
};

#endif // __PITCHDETECTOR_H_
